__title__ = "zrp"
__version__ = "0.1.19"
__release_download_url_prefix__ = "https://github.com/zestai/zrp/archive/refs/tags/"
__download_url_prefix__ = "https://github.com/zestai/zrp/releases/download/"
__projects__ = "https://github.com/zestai"
__projects_branch__ = "main"


